unit OCls;

{**************************************************************}
{ This components work with Oracle Call Interface without BDE. }
{ It is test version and not free of bugs.                     }
{ Author: Albert A. Mavrin                                     }
{ E-mail: amavr@yahoo.com                                      }
{**************************************************************}

interface
uses
  Windows, Messages, SysUtils, Classes, Controls, Forms,
  Db, OciApi;

type
  TMsgLang = (ml_English, ml_Other);
  TFieldState = (fs_Trunc, fs_Exact, fs_Null);

const
  ERR_MSG_LEN = 512;

  ROWID_SIZE = 18;
  FIELD_NAME_LEN = 30;
  DEF_BOOKMARK_SIZE = 10;
  DEF_RECORD_IN_BLOCK = 100;
  DefaultLang: TMsgLang = ml_English;

type
  PInd = ^TInd;
  TInd = sb2;
  PByte = ^Byte;

  PInfoRec = ^TInfoRec;
  TInfoRec = packed record
    BookMarkString: String[DEF_BOOKMARK_SIZE];
    RecordNumber: Integer;
    BookMarkFlag: TBookMarkFlag;
  end;

  TFldInfo = ^FldInfoRec;
  FldInfoRec = packed record
    InBufOffs: Integer;      // Offset into buffer
    InBufSize: Integer;      // Size into buffer
    ColSize: Longint;        // Size of column data
    ColType: sb2;            // Type of data
    ColName: PChar;          // Name of column
    ColLeng: Longint;        // Length of column name
    ColView: Longint;        // Length of data display length
    ColPrec: SmallInt;       // Data precision (common precision)
    ColScal: SmallInt;       // Data scale (after point)
    ColNull: SmallInt;       // Mandatory flag
  end;

  TConnectParam = class(TPersistent)
  private
    FUserName: String;
    FPassword: String;
    FConnect: String;
  published
    property UserName: String read FUserName write FUserName;
    property Password: String read FPassword write FPassword;
    property Connect: String read FConnect write FConnect;
  end;

  EOraError = class(EDataBaseError);
  TOraDataSet = class;

  TErrorEvent = procedure(ErrCode: Integer) of object;

  TSessionEvent = procedure(Params: TConnectParam) of object;

  TOraSession = class(TComponent)
  private
    FHda: THda;
    FLda: TLda;
    FErrMsg: PChar;
    FConnectParam: TConnectParam;
    FActive: Boolean;
    FErrCode: sword;
    FDataSets: TList;
    FBeforeOpen: TSessionEvent;
    FAfterOpen: TSessionEvent;
    FBeforeClose: TSessionEvent;
    FAfterClose: TSessionEvent;
    FOnError: TErrorEvent;
  protected
    procedure SetActive(Value: Boolean);
    procedure InternalConnect;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddDataSet(ODS: TOraDataSet);
    procedure RemoveDataSet(ODS: TOraDataSet);
    procedure Connect;
    procedure Disconnect;
    procedure Commit;
    procedure RollBack;
    procedure ShowError;
    property LDA: TLda read FLda write FLda;
  published
    property Active: Boolean read FActive write SetActive;
    property LogParams: TConnectParam read FConnectParam write FConnectParam;

    property BeforeOpen: TSessionEvent read FBeforeOpen write FBeforeOpen;
    property AfterOpen: TSessionEvent read FAfterOpen write FAfterOpen;
    property BeforeClose: TSessionEvent read FBeforeClose write FBeforeClose;
    property AfterClose: TSessionEvent read FAfterClose write FAfterClose;
    property OnError: TErrorEvent read FOnError write FOnError;
  end;

  TOraDataSet = class(TDataSet)
  private
    FCda: TCda;
    FSession: TOraSession;
    FSql: TStrings;
    FSqlType: Word;
    FErrCode: sword;
    FErrMsg: PChar;
    FFldList: TList;
    FPrmList: TList;
    FTmpBuf: Pointer;
    FBufList: TList;
    FParams: TParams;
    FRowSize: Integer;
    FBlockRow: Integer;
    FRowNum: Integer;  // Current record

    FIsOpen: Boolean;
    FBinded: Boolean;
    FDefined: Boolean;
    FFetched: Boolean;
    FAllFetched: Boolean;

    FUpdateAllowed: Boolean;
// Data buffer support
    procedure BufAddBlock;
    procedure BufAddField(Value: TFldInfo);
    procedure BufClear;
    procedure BufDelete;
    function GetRecPtr(ARow: Integer): Pointer;

// Oracle routines
    procedure CursorBind;
    procedure CursorClose;
    procedure CursorDefine;
    procedure CursorDescribe;
    procedure CursorExecute;
    procedure CursorFetch;
    procedure CursorOpen;
    procedure CursorParse;
    procedure ProcessError;
// Parameters support routines
    procedure CreatePrmList;
    procedure FillPrmList;
    procedure ClearPrmList;
// Property support routines
    procedure SetSql(Value: TStrings);
    procedure SetBlockrecord(Value: Integer);
    procedure CreateParams;

  protected
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    function GetActiveRecordBuffer: PChar;
    procedure ClearCalcFields(Buffer: PChar); override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordCount: Integer; override;
    function GetRecordSize: Word; override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetCanModify: Boolean; override;

    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure InternalHandleException; override;

    procedure InternalOpen; override;
    procedure InternalRefresh; override;
    procedure InternalClose; override;
    procedure InternalFirst;override;
    procedure InternalLast;override;
    procedure InternalEdit; override;
    procedure InternalInsert; override;

    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;

    procedure InternalInitFieldDefs; override;
    procedure InternalSetToRecord(Buffer: PChar); override;

    function IsCursorOpen: Boolean; override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    procedure SetSession(Value: TOraSession);
  published
    property Active;
    property Params: TParams read FParams write FParams;
    property RecordsInBlock: Integer read FBlockRow write SetBlockRecord;
    property Session: TOraSession read FSession write SetSession;
    property SQL: TStrings read FSql write SetSql;
    property UpdateAllowed: Boolean read FUpdateAllowed write FUpdateAllowed;

    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
  end;

procedure Register;
function OraDateToDateTime(const Sour: TOraDate): TDateTime;
procedure DateTimeToOraDate(const Sour: TDateTime; Dest: POraDate);

procedure CursorWait;
procedure CursorDefault;

implementation

{$R *.DCR}

const
  ErrorSetRecordCount: array[TMsgLang] of String =
    ('You can`t change number of block record into not empty buffer!',
     '      !');
  ErrorAddField: array[TMsgLang] of String =
    ('You can`t change field set when cursor fetched!',
     '      !');
  ErrorModifiedSQL: array[TMsgLang] of String =
    ('Do not change SQL-statment of opened DataSet!',
     '  SQL-    ');

procedure Register;
begin
  RegisterComponents('Oracle', [TOraSession, TOraDataSet]);
//  RegisterPropertyEditor(TypeInfo(TStrings), TOraDataSet, 'SQL', TSqlProperty);
end;

procedure DebugMsg(const S: String);
begin
  MessageBox(0, PChar(S), PChar('Debug Message!'), mb_OK);
end;

function OraDateToDateTime(const Sour: TOraDate): TDateTime;
var
  TmpC: Word;
  TmpY: Word;
begin
  TmpC := Sour.oraCentury;
  if TmpC < 100
    then TmpC := TmpC + 100;
  TmpY := Sour.oraYear;
  if TmpY < 100
    then TmpY := TmpY + 100;
  with Sour do
    Result := EncodeDate((TmpC - 100) * 100 + TmpY - 100,
                          oraMonth,
                          oraDay) +
              EncodeTime(oraHour - 1, oraMinute - 1, oraSecond - 1, 0);
end;

procedure DateTimeToOraDate(const Sour: TDateTime; Dest: POraDate);
var
  w1, w2, w3, w4 : word;
begin
  with Dest^ do
    begin
      DecodeDate(Sour, w1, w2, w3);
      oraCentury := (w1 div 100) + 100;
      oraYear := (w1 mod 100) + 100;
      oraMonth := w2;
      oraDay := w3;
      DecodeTime(Sour, w1, w2, w3, w4);
      oraHour := w1 + 1;
      oraMinute := w2 + 1;
      oraSecond := w3 + 1;
    end;
end;

procedure CursorWait;
begin
  Screen.Cursor := crSQLWait;
end;

procedure CursorDefault;
begin
  Screen.Cursor := crDefault;
end;

{----------------------------------------------------------}
{-- TOraSession -------------------------------------------}
{----------------------------------------------------------}
constructor TOraSession.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FConnectParam := TConnectParam.Create;
  FDataSets := TList.Create;
  FActive := False;
  FHda := AllocMem(SizeOf(HDADefArr));
  FLda := AllocMem(SizeOf(LDADefRec));
  FErrMsg := AllocMem(ERR_MSG_LEN);
end;

destructor TOraSession.Destroy;
var
  i: Integer;
begin
  if Active then Disconnect;
  FreeMem(FErrMsg, ERR_MSG_LEN);
  FreeMem(FLda, SizeOf(LDADefRec));
  FreeMem(FHda, SizeOf(HDADefArr));
  for i := 0 to FDataSets.Count - 1 do
    TOraDataSet(FDataSets.Items[i]).FSession := nil;
  FDataSets.Free;
  FConnectParam.Free;
  inherited Destroy;
end;

procedure TOraSession.SetActive(Value: Boolean);
begin
  if Value = FActive then Exit;
  if Value
    then Connect
    else Disconnect;
end;

procedure TOraSession.AddDataSet(ODS: TOraDataSet);
begin
  FDataSets.Add(ODS);
end;

procedure TOraSession.RemoveDataSet(ODS: TOraDataSet);
begin
  FDataSets.Remove(ODS);
end;

procedure TOraSession.InternalConnect;
begin
  if csReading in ComponentState then Exit;
  with FConnectParam do begin
    CursorWait;
    FErrCode := OLog(FLda, FHda,
                     PChar(UserName), -1,
                     PChar(Password), -1,
                     PChar(Connect), -1,
                     OCI_LM_DEF);
    CursorDefault;
  end;
  if FErrCode = 0
    then FActive := True
    else ShowError;
end;

procedure TOraSession.Connect;
begin
  if Assigned(FBeforeOpen)
    then BeforeOpen(FConnectParam);
  if FActive then Exit;
  InternalConnect;
  if Assigned(FAfterOpen)
    then AfterOpen(FConnectParam);
end;

procedure TOraSession.Disconnect;
var
  i: Integer;
begin
  if Assigned(FBeforeClose)
    then BeforeClose(FConnectParam);
  if not FActive then Exit;
  for i := 0 to FDataSets.Count - 1 do
    TDataSet(FDataSets.Items[i]).Close;
  CursorWait;
  FErrCode := OLogof(FLda);
  CursorDefault;
  if FErrCode <> 0
    then ShowError;
  FActive := False;
  if Assigned(FAfterClose)
    then AfterClose(FConnectParam);
end;

procedure TOraSession.Commit;
begin
  if not FActive then Exit;
  CursorWait;
  FErrCode := OCom(FLda);
  CursorDefault;
  if FErrCode <> 0 then ShowError;
end;

procedure TOraSession.RollBack;
begin
  if not FActive then Exit;
  CursorWait;
  FErrCode := ORol(FLda);
  CursorDefault;
  if FErrCode <> 0 then ShowError;
end;

procedure TOraSession.ShowError;
begin
  if Assigned(FOnError)
    then OnError(FErrCode);
  if FErrCode <> 0 then begin
    OErhms(FLda, FErrCode, FErrMsg, ERR_MSG_LEN);
    raise EOraError.Create(String(FErrMsg));
  end;
end;

{----------------------------------------------------------}
{-- TOraDataSet -------------------------------------------}
{----------------------------------------------------------}
{-- Private declaration --}
procedure TOraDataSet.BufAddBlock;
begin
  FBufList.Add(AllocMem(FRowSize * FBlockRow));
end;

procedure TOraDataSet.BufAddField(Value: TFldInfo);
begin
  // If Buffer has allocated memory blocks
  if FBufList.Count > 0
    // then disable change of record size
    then raise Exception.Create(ErrorAddField[DefaultLang]);

  // Set field size into buffer
  FFldList.Add(Value);
  with Value^ do
    case Value.ColType of
      oexFloat:   InBufSize := SizeOf(Double);
      oexInteger: InBufSize := SizeOf(Integer);
      oinRowId:   InBufSize := ROWID_SIZE;
      else InBufSize := ColSize;
    end;
end;

procedure TOraDataSet.BufClear;
begin
  // Delete data blocks
  while FBufList.Count > 0 do begin
    FreeMem(FBufList.Items[FBufList.Count - 1], FRowSize * FBlockRow);
    FBufList.Delete(FBufList.Count - 1);
  end;
end;

procedure TOraDataSet.BufDelete;
var
  i: Integer;
begin
  BufClear;
  // Delete field info
  for i := 0 to FFldList.Count - 1 do begin
    FreeMem(TFldInfo(FFldList.Items[i])^.ColName, FIELD_NAME_LEN);
    FreeMem(FFldList.Items[i], SizeOf(FldInfoRec));
  end;
  // Delete field list
  FFldList.Clear;
  // Reset all sizes and offsets
  FRowSize := 0;
end;

procedure TOraDataSet.CreatePrmList;
var
  i: Integer;
  p: Pointer;
begin
  for i := 0 to Params.Count - 1 do begin
    if Params.Items[i].DataType in [ftDate, ftTime, ftDateTime]
      then p := AllocMem(SizeOf(TOraDate))
      else p := AllocMem(Params.Items[i].GetDataSize);
    FPrmList.Add(p);
  end;
end;

procedure TOraDataSet.FillPrmList;
var
  i: Integer;
  d: TDateTime;
begin
  for i := 0 to Params.Count - 1 do
    if Params.Items[i].DataType in [ftDate, ftTime, ftDateTime]
      then begin
        d := Params.Items[i].AsDateTime;
        DateTimeToOraDate(d, FPrmList.Items[i]);
      end
      else Params.Items[i].GetData(FPrmList.Items[i]);
end;

procedure TOraDataSet.ClearPrmList;
begin
  while FPrmList.Count > 0 do begin
    FreeMem(FPrmList.Items[FPrmList.Count - 1],
            Params.Items[FPrmList.Count - 1].GetDataSize);
    FPrmList.Delete(FPrmList.Count - 1);
  end;
end;

procedure TOraDataSet.SetSql(Value: TStrings);
begin
  if Active
    then raise EOraError.Create(ErrorModifiedSQL[DefaultLang]);
  if SQL.Text <> Value.Text then
  begin
    SQL.BeginUpdate;
    try
      SQL.Assign(Value);
      CreateParams;
    finally
      SQL.EndUpdate;
    end;
  end;
end;

procedure TOraDataSet.CreateParams;
var
  TmpParams: TParams;
begin
  TmpParams := TParams.Create;
  TmpParams.ParseSQL(Sql.Text, True);
  TmpParams.AssignValues(Params);
  Params.Clear;
  Params.Assign(TmpParams);
  TmpParams.Free;
end;

procedure TOraDataSet.SetBlockrecord(Value: Integer);
begin
  if FBufList.Count > 0
    then raise Exception.Create(ErrorSetRecordCount[DefaultLang]);
  FBlockRow := Value;
end;

procedure TOraDataSet.CursorBind;
var
  i: Integer;
  ValPtr: Pointer;
  Ind: TInd;
  IndPtr: Pointer;
  ODataType: sword;
  ODataSize: sword;
begin
  for i := 1 to Params.Count do begin
    ValPtr := FPrmList.Items[i - 1];
    case Params.Items[i - 1].DataType of
      ftUnknown,
      ftString : begin
                   ODataType := oexVarChar2;
                   ODataSize := -1;
                 end;

      ftSmallint,
      ftWord,
      ftBoolean : begin
                     ODataType := oexInteger;
                     ODataSize := SizeOf(Word);
                  end;
      ftInteger : begin
                    ODataType := oexInteger;
                    ODataSize := SizeOf(Integer);
                  end;

      ftFloat,
      ftCurrency,
      ftBCD : begin
                ODataType := oexFloat;
                ODataSize := SizeOf(Double);
              end;

      ftDate,
      ftTime,
      ftDateTime : begin
                     ODataType := oexDate;
                     ODataSize := SizeOf(TOraDate);
                   end;

      else begin
        ODataType := oexLong;
        ODataSize := -1;
      end;
    end;
    if Params.Items[i - 1].IsNull
      then Ind := -1
      else Ind := 0;
    IndPtr := @Ind;
    FErrCode := OBndrv(FCda,
                       PChar(':' + Params.Items[i - 1].Name),
                       -1,
                       ValPtr,
                       ODataSize,
                       ODataType,
                       -1,
                       IndPtr, nil, -1, -1);
  end;
  FBinded := True;
end;

procedure TOraDataSet.CursorClose;
begin
  if FDefined
    then begin
      FreeMem(FTmpBuf, FRowSize * FBlockRow);
      FDefined := False;
    end ;
  ClearPrmList;
  BufDelete;
  Screen.Cursor := crSQLWait;
  FErrCode := OClose(FCda);
  Screen.Cursor := crDefault;
  if FErrCode <> 0
    then ProcessError;
  FIsOpen := False;
end;

procedure TOraDataSet.CursorDefine;
var
  i: Integer;
  BufOfs: PChar;
begin
  if not FDefined then begin
    FTmpBuf := AllocMem(FRowSize * FBlockRow);
    FDefined := True;
  end;

  CursorWait;
  for i := 0 to FFldList.Count - 1 do begin
    BufOfs := PChar(FTmpBuf) + TFldInfo(FFldList.Items[i])^.InBufOffs;
    FErrCode := ODefinps(FCda,
                         1,
                         i + 1,
                         BufOfs + SizeOf(TInd),
                         TFldInfo(FFldList.Items[i])^.InBufSize,
                         TFldInfo(FFldList.Items[i])^.ColType,
                         0,
                         BufOfs,
                         nil, 0, 0, nil, nil,
                         FRowSize,
                         FRowSize,
                         0, 0);
    if FErrCode <> 0 then Break;
  end;
  CursorDefault;
  if FErrCode <> 0 then begin
    FreeMem(FTmpBuf, FRowSize * FBlockRow);
    FDefined := False;
    ProcessError;
  end;
end;

procedure TOraDataSet.CursorDescribe;
var
  i: Integer;
  P: PChar;
  ColDesc: TFldInfo;
begin
  i := 0;
  CursorWait;
  repeat
    Inc(i);

    ColDesc := AllocMem(SizeOf(FldInfoRec));
    ColDesc^.ColName := AllocMem(FIELD_NAME_LEN);
    ColDesc^.ColLeng := FIELD_NAME_LEN;

    FErrCode := ODescr(FCda, i,
                       @ColDesc^.ColSize,
                       @ColDesc^.ColType,
                        ColDesc^.ColName,
                       @ColDesc^.ColLeng,
                       @ColDesc^.ColView,
                       @ColDesc^.ColPrec,
                       @ColDesc^.ColScal,
                       @ColDesc^.ColNull);
    if FErrCode = 0
      then with ColDesc^ do begin
        if ColLeng < FIELD_NAME_LEN then begin
          P := ColName + ColLeng;
          P^ := #0;
        end;
        case ColType of
          oinNumber   : ColType := oexFloat;
          oinRowId    : begin
                          ColSize := ROWID_SIZE;
                          ColType := oexVarChar2;
                        end;
        end;
      end
      else begin
        FreeMem(ColDesc^.ColName, FIELD_NAME_LEN);
        FreeMem(ColDesc, SizeOf(FldInfoRec));
        if FCda.rc = oerrVarNotInList
          then Break;
      end;

    BufAddField(ColDesc);

    // Set row size = (data size + indicator size)
    // and field offset into buffer
    ColDesc^.InBufOffs := FRowSize;
    Inc(FRowSize, ColDesc^.InBufSize + SizeOf(TInd));
  until FErrCode <> 0;
  CursorDefault;
  if FCda.rc <> oerrVarNotInList
    then ProcessError;
end;

procedure TOraDataSet.CursorExecute;
begin
  FillPrmList;
  CursorWait;
  FErrCode := OExec(FCda);
  CursorDefault;
  if FErrCode <> 0
    then ProcessError;
end;

procedure TOraDataSet.CursorFetch;
var
  Buf: Pointer;
begin
  CursorWait;
  FErrCode := OFen(FCda, FBlockRow);
  CursorDefault;
  FFetched := FCda.rpc > 0;
  if ((FErrCode = 0) or (FCda.rc = oerrNoDataFound)) and FDefined
    then begin
      BufAddBlock;
      Buf := FBufList.Items[FBufList.Count - 1];
      Move(FTmpBuf^, Buf^, FRowSize * FBlockRow);
      FillChar(FTmpBuf^, FRowSize * FBlockRow, 0);
      FAllFetched := FCda.rc = oerrNoDataFound;
    end;
  if FCda.rc <> oerrNoDataFound
    then ProcessError;
end;

procedure TOraDataSet.CursorOpen;
begin
  CursorWait;
  FErrCode := OOpen(FCda, FSession.Lda, nil, -1, -1, nil, -1);
  FAllFetched := False;
  FFetched := False;
  FDefined := False;
  FBinded := False;
  FRowNum := -1;
  CursorDefault;
  CreatePrmList;

  FIsOpen := FErrCode = 0;
  if not FIsOpen
    then ProcessError;
end;

procedure TOraDataSet.CursorParse;
begin
  CursorWait;
  FErrCode := OParse(FCda, PChar(FSql.Text + #0), -1, 0, VERSION_7);
  CursorDefault;

  ProcessError;

  FSqlType := FCda.ft;
  if FDefined
    then begin
      FreeMem(FTmpBuf, FRowSize * FBlockRow);
      FDefined := False;
    end;
end;

function TOraDataSet.GetRecPtr(ARow: Integer): Pointer;
var
  I: Integer;
begin
  if (FBufList.Count > 0) and
     (ARow >= 0) and
     (ARow < FBufList.Count * FBlockRow)
    then begin
      // get block number
      I := ARow div FBlockRow;
      // get block address
      Result := FBufList.Items[I];
      // get record position into block
      I := ARow - I * FBlockRow;
      // add offset for result
      Inc(PChar(Result), FRowSize * I);
    end
    else Result := nil;
end;

procedure TOraDataSet.ProcessError;
begin
  if FCda.rc = 0 then Exit;
  OErhms(FSession.FLda, FErrCode, FErrMsg, ERR_MSG_LEN);
  raise EOraError.Create(String(FErrMsg));
end;

{-- Protected declaration --}
function TOraDataSet.AllocRecordBuffer: PChar;
begin
  // get memory for record buffer
  Result := AllocMem(GetRecordSize);
end;

procedure TOraDataSet.FreeRecordBuffer(var Buffer: PChar);
begin
  // free memory of record buffer
  FreeMem(Buffer, GetRecordSize);
end;

function TOraDataSet.GetActiveRecordBuffer: PChar;
begin
  case State of
    dsBrowse: if not FFetched
                then Result := nil
                else Result := ActiveBuffer;

    dsCalcFields: Result := CalcBuffer;

    dsFilter, dsEdit, dsInsert: Result:= ActiveBuffer;

    else Result:=nil;
  end;
end;

function TOraDataSet.GetRecordCount: Integer;
begin
  if FSqlType = SQL_SELECT
    then Result := FCda.rpc
    else Result := 0;
end;

function TOraDataSet.GetRecordSize: Word;
begin
  Result := FRowSize
          + SizeOf(TInfoRec)
          + CalcFieldsSize;
end;

function TOraDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  RecPtr: Pointer;
begin
  if not FFetched
  then begin
    Result := grEOF;
    Exit;
  end;
  Result := grOk;

  case GetMode of

    gmCurrent: if FRowNum >= RecordCount
                 then Result:=grEOF
                 else if FRowNum < 0
                        then Result:=grBOF
                        else Result:=grOk;

    gmNext: if FRowNum >= RecordCount - 1
              then if FAllFetched
                     then Result := grEOF
                     else CursorFetch
              else Inc(FRowNum);

    gmPrior: if FRowNum > 0
               then Dec(FRowNum)
               else Result := grBOF;
  end;

  if Result = grOk
    then begin
      RecPtr := GetRecPtr(FRowNum);
      Move(RecPtr^, Buffer^, FRowSize);
      with PInfoRec(Buffer + FRowSize)^ do begin
        RecordNumber := FRowNum;
        BookMarkString := '';
        BookMarkFlag := bfCurrent;
      end;
    end;

  ClearCalcFields(Buffer);
  GetCalcFields(Buffer);
end;

procedure TOraDataSet.ClearCalcFields(Buffer: PChar);
var
  P: PChar;
begin
  P := Buffer + FRowSize + SizeOf(TInfoRec);
  FillChar(P^, CalcFieldsSize, 0);
end;

function TOraDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PInfoRec(Buffer + FRowSize)^.BookMarkFlag;
end;

function TOraDataSet.GetCanModify: Boolean;
begin
//  Result := UpdateAllowed;
  Result := False;
end;

procedure TOraDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  Move((Buffer + FRowSize)^, Data^, DEF_BOOKMARK_SIZE);
end;

procedure TOraDataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PInfoRec(Buffer + FRowSize)^.BookMarkFlag := Value;
end;

procedure TOraDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  Move(Data^, (Buffer + FRowSize)^, DEF_BOOKMARK_SIZE);
end;

procedure TOraDataSet.InternalHandleException;
begin
  DebugMsg('InternalHandleException');
end;

procedure TOraDataSet.InternalOpen;
begin
  if not Assigned(FSession) then Exit;
  FSession.Connect;
  if not FSession.Active then Exit;
  CursorOpen;
  CursorParse;
  CursorBind;
  if FSqlType = SQL_SELECT then begin
    CursorDescribe;
    InternalInitFieldDefs;
    if DefaultFields then CreateFields;
    CursorDefine;
    CursorExecute;
    CursorFetch;
    BindFields(True);
  end
  else
    CursorExecute;
end;

procedure TOraDataSet.InternalRefresh;
begin
  BufDelete;
  CursorExecute;
  if FSqlType = SQL_SELECT then begin
    CursorFetch;
    BindFields(True);
  end;
end;

procedure TOraDataSet.InternalClose;
begin
  BindFields(False);
  if DefaultFields then DestroyFields;
  CursorClose;
end;

procedure TOraDataSet.InternalFirst;
begin
  FRowNum := - 1;
end;

procedure TOraDataSet.InternalLast;
begin
  if FSqlType <> SQL_SELECT then Exit;
  while not FAllFetched do
    CursorFetch;
  FRowNum := RecordCount;
end;

procedure TOraDataSet.InternalEdit;
begin
  DebugMsg('InternalEdit');
end;

procedure TOraDataSet.InternalInsert;
begin
  DebugMsg('InternalInsert');
end;

procedure TOraDataSet.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, GetRecordSize, 0);
  PInfoRec(Buffer + FRowSize)^.RecordNumber := GetRecordCount;
  PInfoRec(Buffer + FRowSize)^.BookMarkString := '';
  PInfoRec(Buffer + FRowSize)^.BookMarkFlag := bfCurrent;
end;

procedure TOraDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  DebugMsg('InternalAddRecord');
end;

procedure TOraDataSet.InternalInitFieldDefs;
var
  i:   smallint;
  FldSize: Word;
  FldType: TFieldType;
  FldName: String;

begin
  FieldDefs.Clear;

  if not IsCursorOpen
    then begin
      FSession.Connect;
      CursorOpen;
      CursorParse;
      CursorBind;
      CursorDescribe;
    end;

  for i := 0 to FFldList.Count - 1 do begin
    FldSize := 0;
    SetString(FldName,
              TFldInfo(FFldList.Items[i])^.ColName,
              TFldInfo(FFldList.Items[i])^.ColLeng);
    case TFldInfo(FFldList.Items[i])^.ColType of
      oexRaw,
      oexChar,
      oexVarChar2 : begin
                      FldType := ftString;
                      FldSize := TFldInfo(FFldList.Items[i])^.ColSize;
                    end;
       oexInteger : begin
                      FldType := ftInteger;
                    end;
         oexFloat : begin
                        FldType := ftFloat;
                      end;
       oexLong,
       oexLongRaw : begin
                      FldType := ftBlob;
                      FldSize := SizeOf(Word) * 4;
                    end;
          oexDate : FldType := ftDateTime;

      else FldType := ftUnknown;

    end;

    with TFieldDef.Create(FieldDefs, FldName,
                          FldType, FldSize,
                          TFldInfo(FFldList.Items[i])^.ColNull = 0, i) do
      InternalCalcField := False;
  end;
end;

procedure TOraDataSet.InternalSetToRecord(Buffer: PChar);
begin
  FRowNum := PInfoRec(Buffer + FRowSize)^.RecordNumber;
end;

function TOraDataSet.IsCursorOpen: Boolean;
begin
  Result := FIsOpen;
end;

procedure TOraDataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
  SourceBuffer: PChar;
begin
  if (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup)
    then begin
      SourceBuffer := GetActiveRecordBuffer;
      Inc(SourceBuffer, FRowSize + SizeOf(TInfoRec));
      Inc(SourceBuffer, Field.Offset);
      Move(Buffer^, SourceBuffer^, Field.DataSize);
    end;
end;

{-- Public declaration --}
constructor TOraDataSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSql := TStringList.Create;
  FCda := AllocMem(SizeOf(CDADefRec));
  FParams := TParams.Create(Self);
  FFldList := TList.Create;
  FPrmList := TList.Create;
  FBufList := TList.Create;
  FDefined := False;
  FUpdateAllowed := False;
  FRowSize := 0;
  FBlockRow := DEF_RECORD_IN_BLOCK;
  FAllFetched := False;
  FIsOpen := False;
  FErrMsg := AllocMem(ERR_MSG_LEN);
end;

destructor TOraDataSet.Destroy;
begin
  if Active then Close;
  FreeMem(FErrMsg, ERR_MSG_LEN);
  FPrmList.Free;
  FFldList.Free;
  FBufList.Free;
  if Assigned(FSession) then
    FSession.FDataSets.Remove(Self);
  FParams.Free;
  FreeMem(FCda, SizeOf(CDADefRec));
  FSql.Free;
  inherited Destroy;
end;

function TOraDataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  SourceBuffer: PChar;
  DTTmp: TDateTime;
  StrTmp: String;
begin
  Result := False;
  SourceBuffer := GetActiveRecordBuffer;

  if not FIsOpen or (SourceBuffer = nil)
    then Exit;

  if (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup)
    then begin
      // set pointer to calc buffer
      Inc(SourceBuffer, FRowSize + SizeOf(TInfoRec));
      // set pointer to address of field data
      Inc(SourceBuffer, Field.Offset);
      if (SourceBuffer^ = #0) or (Buffer = nil)
        then Exit
        // copy field data to buffer
        else Move(SourceBuffer^, Buffer^, Field.DataSize);
    end
    else begin
      // Record buffer structure is:
      // |IND(2)|FIELD_DATA(field size)|IND(2)|FIELD_DATA(field size)....
      // Set SourceBuffer to needing field into record buffer
      Inc(SourceBuffer, TFldInfo(FFldList.Items[Field.FieldNo - 1])^.InBufOffs);

      // if field data is null
      if PInd(SourceBuffer)^ < 0
        then Exit
        else Inc(SourceBuffer, SizeOf(TInd));
      case Field.DataType of

        ftDateTime: begin
            DTTmp := OraDateToDateTime(POraDate(SourceBuffer)^);
            Double(Buffer^) := TimeStampToMSecs(DateTimeToTimeStamp(DTTmp));
        end;

        ftString:
        begin
          StrTmp := TrimRight(Copy(PChar(SourceBuffer), 1, Field.DataSize - 1));
          FillChar(Buffer^, TFldInfo(FFldList.Items[Field.FieldNo - 1])^.ColSize, 0);
          StrLCopy(PChar(Buffer), PChar(SourceBuffer), Length(StrTmp));
        end;

        else
          Move(SourceBuffer^, Buffer^,
               TFldInfo(FFldList.Items[Field.FieldNo - 1])^.InBufSize);
      end;
    end;
  Result:=True;
end;

procedure TOraDataSet.SetSession(Value: TOraSession);
begin
  if FSession = Value then Exit;
  if Assigned(FSession) then
    FSession.RemoveDataSet(Self);
  if Assigned(Value) then
    Value.AddDataSet(Self);
  FSession := Value;
end;

end.
